// ================================================================
// Morse Code Trainer   https://www.instructables.com/id/Morse-Code-Trainer/
// Currently using: STEM / Educational Game Platform - using a top level Menu selection
// with a Micro Controller, 8 LED 4 color bar and 4 Button module + piezo sounder
//
// license: Attribution-NonCommercial-ShareAlike
// Jan-July 2020 Ron Miller, Initial Development

//  include Library of Hardware Support functions
#include "4Btn_8LED_Platform.h"

//#define DEBUG false
bool DEBUG = false;
const int BEEPPIN = 13;  // to passive piezo sounder
#define BUZZER_ON 1
#define BUZZER_OFF 0

// ------------------------ function prototypes
void quickCalib();
void echoMorse();
void A2Z_Morse_Exercise();
void Morse_AB_Exercise();
void Relay_exercise();
void Tx_10Word_test();
void Rx_10Word_test();

byte fetchMorseChar();
char dida2Ascii(byte);
byte Ascii2dida(char);
void set_didaStr(byte, char[]);

void dit_dah(unsigned short int );
void ditAudio(unsigned short int);
void dahAudio(unsigned short int);
void playMorseCode(byte, int);
void playStr(char*);

void showLevel(byte);
byte get_Option(byte, int);
void ranIndexes(byte[], byte);
bool btnDown();
void stableBtns();

unsigned long t0,t1;
unsigned int avgPeriod, pAveraging;
unsigned int ditPeriod;
int farns; // Farnsworth Factor, multiplier for std to Farnsworth inter char & word periods

bool oneKey=true; // straight single contact key   (TBD: using a 2 contact paddle)
bool batOp=false; // would-be indicator for battery Operation (w/o serial monitor)
char didaStr[10];
unsigned long randSeed=0;
const char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";
const char wordSet[6][62] = { 
      "The quick brown fox jumps over the lazy dog. Wowzer!\0",
      "Pack my box with five dozen liquor jugs. You betcha.\0",
      "The five boxing wizards jump quickly. OMG, what a show.\0",
      "Bad Boys Run Over Your Garden, But Violets Gone Wild.\0",
      "Cozy lynx gives smart squid who asks for job pen.\0",    // found this late, & changed 'lummox' to 'lynx'
      "Now, Vicky forged my jewelry next to the boutique plazas.\0"  };  // with +3 words over line below for 10
    // Vicky forged handsome boutique plaza ax jewelry.  // my original creation
    // ideas?... CQCQ KI7NEW SQO, QRS 15 QRN? R5S9, SOS QTH 73.

const char nonAlpha[] = "~1234567890.,?/'-:\"+!=@$)&\n";
                          //        1    2    3    4    5    6    7    8    9    0
const byte nonAlpha_dida[] = {0,0x10,0x18,0x1C,0x1E,0x1F,0xEF,0xE7,0xE3,0xE1,0xE0,
       0x2A,0xCC,0x33,0xED,0x21,0xDE,0xC7,0x2D,0x15,0xD4,0xEE,0x25,0x76,0xD2,0x17,0x0A};
     // .    ,    ?    /    '    -    :    "    +    !    =    @    $    )    &    \n

// the setup routine runs once on power up or when you press reset:
//=================================================================
void setup() {

  // ------------ setup the Mini Game/STEM platform (8LEDs, 4Btns, and Audio)
  gPlatform_setup();

  // initialize serial communication
  Serial.begin(115200);
  Serial.println("----------------------- Morse Code Trainer");

  // ------ setup optional behavior based on button states at startup
  stableBtns(); // get btns state after debouncing
  if (Btn) {
    // ahead btn1 held at startup disables Farnsworth
    if (btn2) oneKey = false;  // btn2 for 2-Key, single contact straight key (1 button) by default
    if (btn3) randomSeed(randSeed=314159); // always use the same rand #s for each run
    DEBUG = btn4;
    Serial.print("button state at startup: "); 
    Serial.print(btn1); Serial.print(btn2); Serial.print(btn3); Serial.println(btn4);
  }

  ditPeriod= (60000)/(240*2);  // in msecs, for ~10WPM to start off  (240 in 60,000 ms is 5WPM)
//  ditPeriod=(8*ditPeriod/10);  // nah, lets go faster for 12.5 WPM
  avgPeriod= 2*ditPeriod;
  pAveraging= 4*avgPeriod;
  
  // Farnsworth calculations
  int cWPM = 18, sWPM = 10;
  float fp = 1000 * (60/sWPM - 37.2/cWPM) / 19; // inter time period of Farnsworth timing
  farns = fp/(1000*1.2/cWPM);  // factor to convert std to Farnsworth unit time period
  if (btn1) {
    farns = 1; // disable Farnsworth effect if btn1 held down at power up.
    Serial.println("Standard inter-char & word timing used.");
  } else {
    Serial.print("Farnsworth ratio: "); Serial.print(sWPM); Serial.print("/"); Serial.print(cWPM);
    if (DEBUG) {Serial.print("   unit period factor: "); Serial.println(farns);}
    else Serial.println();
  }

// buZZ(30,200);
  delay(200);
  //  demonstrate there is Life in the components (LEDs)
  for (int i=0;i<4;i++) { refreshWait(100); set_LEDs(1<<i,1); set_LEDs(0x80>>i,1);}
  for (int i=4;i<8;i++) { refreshWait(100); set_LEDs(1<<i,2); set_LEDs(0x80>>i,2);}
  for (int i=0;i<4;i++) { refreshWait(100); set_LEDs(1<<i,3); set_LEDs(0x80>>i,3);}
  for (int i=4;i<8;i++) { refreshWait(100); set_LEDs(1<<i,0); set_LEDs(0x80>>i,0);}
  refreshWait(500);
  clearDisp();
  
  // highlight the Audio
  boop(30,200);
  beep(30,200);
  buZZ(30,200);
//  delay(2000);
}

// the loop routine runs over and over again forever:
//=================================================================
void loop() { // Top Menu level, for selecting 1 of N activities
  byte func, maxFunc;

  func=1;
  while(1) {
    clearDisp();
    while(Btn) {buZZ(1); scanBtns();}  // wait for no button

    maxFunc=6;  // currently there are 6 activities/games supported
    // indicate selectable functions 1-6
    for (int i=1; i<=maxFunc; i++) {dim[i]=true;}
    brightOne=func;
    refreshDisp();
    
    // watch for btn presses other than btn1
    while (! btn1) {
      refreshDisp();
      delay(1);
      scanBtns();
      //if (btn2Pressed) brightOne = (brightOne==1)?  maxFunc : brightOne-1;  // -- move left
      if (btn2Pressed) brightOne = (brightOne==0)?  maxFunc : brightOne-1;  // -- move left Allow 0:debugging
      if (btn3Pressed) brightOne = (brightOne%maxFunc) + 1; // ++ moves right
//      if (btn4Pressed) brightOne = (brightOne%maxFunc) + 1; // allow most play with only 2 buttons (1&4)
      if (btn4Pressed) batOp=true;
    }

    // randomize driven by real-world user event
    if (randSeed==0) {
      randomSeed(randSeed = micros()/4);  // micros() resolution is ~ 4us
      // used for QA testing: randomSeed(randSeed=36256276);  // starts with Rx test Ln=8 (7 w/ opt<3)
      Serial.print("random Seed:  "); Serial.print( randSeed); 
    }

    func = brightOne;
    clearDisp();
    while(Btn) {refreshWait(10); scanBtns();} // wait for button to clear

    Serial.print("\n FUNC: "); Serial.println(func);
    switch (func) {
      case 0:
        //btnReporting();    // exercises the buttons and Leds, and reports via Serial
        quickCalib();   // immediate speed calibration
        func=1;
        break;
      case 1:
        // do a series of 'C' characters so it learns your general speed, prior to all other activities
        echoMorse();
        break;
      case 2:
        A2Z_Morse_Exercise();   // key in & lisen to letters A to Z one letter at a time
        break;
      case 3:
        Morse_AB_test();     // given a letter of the alphabet key in the one that follows it
        break;
      case 4:
        //printTest_Chr_Ditdah();
        Relay_exercise();   // copy a morse code pseudo word and then key it in
        break;
      case 5:
        Tx_10Word_test();   // Test: type out 10 word sentence & get WPM & Err cnt 
        break;
      case 6:
        Rx_10Word_test();   // Test: Copy down 10 words at desired speed. Self scoring.
        break;
      default:
        buZZ(1000);
    }
  }
}


//================== major project funtional building blocks ===================

byte fetchMorseChar(bool echo) {
  if (oneKey) return(fetchMorse_1Key(echo));  // uses One button normalized keying
  else        return(fetchMorse_2Key(echo));  // uses Two button normalized keying
}

// ---------- Morse code support functions ------------
byte fetchMorse_1Key(bool echo) { // get one char (ditdah set) using single button keying
  byte mChar=0;
  int  p=0;
  unsigned long pt;
  byte dit;
  bool toggle=false;
  static bool avgdit = true;

  // update timing used prior to each whole character
  avgPeriod = pAveraging/4; // the running average maintains 4*resolution desired
  ditPeriod=avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   
//Serial.print(avgPeriod);  Serial.print("  ");

  while(!(btn1 || btn3 || btn4)) {scanBtns(); refreshWait(1);} // wait for button press
  while (btn1) {
    p=0;
    pt = millis();  // time the period
    while (btn1) {
      dit_dah(1);   // output audio while key pressed
      scanBtns();
    }
    p = millis()-pt;
    
    clearDisp();
//    dit = (p<avgPeriod);
    dit = (p<(9*avgPeriod/10));  // makes 1.8 ditPeriod the break point between a dit & a dah
    if (mChar==0) { // setup first bit
      mChar = (dit)? 0x01 : 0xFE;
    } else {    // shift in other bits
      mChar = (mChar<<1) | dit;
    }
    // add into the running average 'pAveraging' (but not ones from left field)
    if (avgdit && dit && (2*p>ditPeriod)) {
      pAveraging = (3*pAveraging/4) + 2*p;
      avgdit = !avgdit; // average in = # of dits & dahs
    }
    if (!avgdit && !dit && p<(5*ditPeriod)) {
      pAveraging = (3*pAveraging/4) + 2*(p/3);
      avgdit = !avgdit;
    }
    
    p=0;
    while (Btn==0) {  // wait for additional dit-dahs (or other)
      scanBtns();
      delay(1);
//      if (++p>(7*ditPeriod/3)) {  // require only 2.3 (3 expected be the average)
      if (++p>(2*ditPeriod)) {
        break;          // done with this char
      }
    }
  }
  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  Serial.print(dida2Ascii(mChar));  // always output the alpha character
  if (DEBUG) {
    Serial.print("   Morse Char: ");  Serial.print(mChar);
    Serial.print("  0x"); Serial.print(mChar, HEX);
    Serial.print("    ave Period: ");  Serial.println(pAveraging/4);
  }

  return (mChar);
}

// -----------------------   
byte fetchMorse_2Key(bool echo) { // using single button keying
     // not throurghly valided, may need adjustments to period timing
  static short int ditPeriod;
  byte mChar=0;
  int  p=0, p2=0, pCnt=0;
  bool dit;
  byte msn;
  bool toggle=false;
  byte ditdaCnt=0;

  avgPeriod = pAveraging/4;   // the running average maintains 4*resolution desired
  ditPeriod = avgPeriod/2;    // avgPeriod=2*ditPeriod  aka  the average between dit and dah period   
  
  while(!Btn) {scanBtns(); refreshWait(1);} // wait for button press
  while (btn1 || btn2) {
    ditdaCnt++;
    p+=p2;
    dit = btn1;
    while (Btn) {
      scanBtns();
        if (btn1) {   // produce sound while pressing
          brightOne=7;
          ditAudio(1);
        } else {
          brightOne=7; litLED=8;
          dahAudio(2);
        }
      toggle = !toggle;
      p++;
      clearDisp();
    }

    if (mChar==0) { // setup first bit
      mChar = (dit)? 0x01 : 0xFE;
    } else {    // shift in other bits
      mChar = (mChar<<1) | dit;
    }
    // count it into the running average 'pAveraging'
    if (dit) pCnt+=2;
    else pCnt+=3;   // compromise, as it takes 2 to key but 4 to reproduce
    
    p2=0;
    while (Btn==0) {  // wait for additional dit-dahs
      scanBtns();
      delay(1);
      if (++p2>(2*ditPeriod)) {
        break;      // done here, end of a char
      }
    }
  }

  if (ditdaCnt>1) {
    ditPeriod = p/pCnt;
//    pAveraging = (3*pAveraging/4) + 2*(ditPeriod);  // provide tracking of avg. period of a ditdah
    pAveraging = (3*pAveraging/4) + 3*(ditPeriod);  // partially arbitarry & empirical
  }
  
  if (btn3 || btn4) return 0;
  if (echo) {
    set_didaStr(mChar, didaStr);
    Serial.print("  ");
    Serial.print(didaStr); Serial.print(" ");
  }
  Serial.print(dida2Ascii(mChar));  // always output the alpha character

  if (DEBUG) {
    Serial.print("   Morse Char: ");  Serial.print(mChar);
    Serial.print("  0x"); Serial.print(mChar, HEX);
    Serial.print("    ave Period: ");  Serial.println(2*ditPeriod);
    Serial.print(" Traking Period: ");  Serial.println(pAveraging/4);
  }

  return (mChar);
}

// -----------------------   
char dida2Ascii(byte ditdahs) {
  byte ncode, ascii, msb, ndidas, i;
  const char nc2A[]  = "~EAIWRUSJP\nL~FVH~~QZYCXBOGKDMNT~";
  const char dida5[] = "09 8   7     /=61    + &2   3 45";
  const char dida6[] = " '   @ :  . ,\"    )?! $       - ";

  if (ditdahs==0) return 0;

  // strip off all most significant extra 1s bits
  i=1;
  msb = (ditdahs & 0x80);
  while ((ditdahs<<i & 0x80) == msb) { i++;}
  ndidas = 8-i;
  ncode  = ditdahs & 0x1F;
  // decode according to its length
  if (ndidas<5)       ascii = nc2A[ncode];
  else if (ndidas==5) ascii = dida5[ncode];
  else if (ndidas>5)  ascii = dida6[ncode];

  return (ascii);
}

// -----------------------   
byte Ascii2dida(char Achar) {
  byte ncode, i;
  const char nc2A[] = "~EAIWRUSJP~L~FVH~~QZYCXBOGKDMNT~~";
  int alen;

  Achar = toupper(Achar);
  //  move i-index past padding bits
  for (i=1; i<32; i++) {if (nc2A[i]==Achar) break;}
  if (i<32) {
    ncode = (i<B10000) ? i : i+B11100000;
  } else {  // handle #s, punc. & spec. chars
    alen = sizeof(nonAlpha);  // sizeof(char)=1
    for (i=1; i<alen; i++) {if (nonAlpha[i]==Achar) break;}
    ncode = nonAlpha_dida[i];
  }
  return (ncode);
}

// --------------------------------------------   
void set_didaStr(byte ditdahs, char ddStr[]) {  // string returned in global var: 'didaStr'
  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  byte i=1;
  byte bitState;

  if (ditdahs==0) return;

  while ((ditdahs<<i & 0x80) == (ditdahs & 0x80)) i++;
  int c=0;
  for (int b=i; b<8; b++) {
    if (ditdahs<<b & 0x80) {  // this bit is a 'dit'
      ddStr[c++]='.';    
    } else {  // its a dah
      ddStr[c++]='-';    
    }
  }
  ddStr[c++]=0;
}

//================= Main Morse Code Activities / Applets =================

// -----------------------   
void quickCalib() {
  byte dida;
  bool echo=true;
  unsigned long t;
  float wpm;
  
  while(!btn4) {
    clearDisp();
    refreshWait(1); 
    if (btnPressed()) {    // Button has been pushed
      if (btn1) {
        //  - - - -  time four dits
        t=0;
        t0=millis();
        while(btn1) {dit_dah(1); scanBtns();}
        t+=millis()-t0;               // use just the dit periods,
        while(!btn1) scanBtns();      // because the pauses can be unpredictable
        t0=millis();
        while(btn1) {dit_dah(1); scanBtns();}
        t+=millis()-t0;
        while(!btn1) scanBtns();
        t0=millis();
        while(btn1) {dit_dah(1); scanBtns();}
        t+=millis()-t0;
        while(!btn1) scanBtns();
        t0=millis();
        while(btn1) {dit_dah(1); scanBtns();} 
        t+=millis()-t0;
        delay(333);
        ditPeriod=t/4;  // set the output rate accordingly
        pAveraging=8*ditPeriod;
        if (ditPeriod<300) playStr("ants");
        delay(500);
        wpm=(5.0*(4.0)/240.0)/(t/1000.0/60.0);  // 5 wpm for each 240*p
        //Serial.print(" avg. Period: "); Serial.println(pAveraging/4);
        Serial.print(" Keying rate (in WPM): "); Serial.println(wpm);
        showLevel(wpm+0.5);
        refreshWait(2000);
        clearDisp();
      }
      if (btn3) {
        wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0)); 
        //Serial.print(" avg. Period: "); Serial.println(pAveraging/4);
        Serial.print(" Keying rate (in WPM): "); Serial.println(wpm);
        playStr("ants");
        showLevel(wpm+0.5);
        refreshWait(2000);
        clearDisp();
      }
    }
  }
}

// -----------------------   
void echoMorse() {
  // do a series of 'C' characters so it learns your general speed
  byte ditdahs;  // all MSBits that are the same are discarded, what's left is 1=dit 0=dah, MSbit is first in time, LSB last
  bool echo=true;
  int chrs;
  unsigned long t, pauseT;
  float wpm;
  int litCnt=0;
  
  // display & btn press is cleared prior
  
  // ------ Echo Morse Code
  while(1) {
    if (btnPressed()) {    // Button has been pushed
      if ((!echo) && pauseT!=0 && (millis()-pauseT)>(4*ditPeriod)) Serial.print(' ');
      if (btn3) {
        if(echo) {
          t0=t1=0;
          chrs=0;
          pauseT=0;
          litCnt=0;
        } else {
          t=t1-t0;
          wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0)); 
          Serial.print("\n current Keying rate: "); Serial.print(wpm);
          //Serial.print("\nchrs: "); Serial.print(chrs); Serial.print("   t: "); Serial.println(t);
          // note: avg char code takes 12p, 4 char is 48p, 5* that is 240p, the std for 5WPM
          wpm = (chrs/4.0)/(t/(1000.0*60.0));
          if (wpm<1) wpm=1;
          Serial.print("    Words/Minute: "); Serial.println(wpm);
          showLevel(wpm+0.5);
          refreshWait(3000); clearDisp();
        }
        echo= !echo;
      }
      
      if (btn1 || btn2) {
        if (t0==0) t0=millis();
        ditdahs = fetchMorseChar(echo);   // get dits & dahs appearing to form the MC of a character
        t1 = millis();
        if (echo) Serial.println();
        pauseT=millis();
        if (ditdahs != 0) {
          chrs++;
          // Echo 
          if (echo) {delay(330);  playMorseCode(ditdahs, 0);}
          else litLED=litCnt=litCnt%8+1;
        } else {
          while(btnDown()) delay(1);
        }
      }
    }
    if (ESC || btn4) break;
    refreshWait(1);
  }
}

// -----------------------   A-Z Morse Code Please
void A2Z_Morse_Exercise() {
//  char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";
  byte ditdahs;  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  byte dida;  // expected
  char cinput;
  byte correct=0;
  static bool echo=true;

  // display & btn press is cleared prior
  delay(100);
  for (int i=1; i<=26; i++) {
    Serial.println(); Serial.print(alpha[i]);
    while(1) {
      delay(1);
      if (btnPressed() && (btn1 || btn2)) {    // Button has been pushed
          ditdahs = fetchMorseChar(true);     // get one char response
          delay(200);  // ?

          cinput = dida2Ascii(ditdahs);
          dida = Ascii2dida(alpha[i]);
          if (cinput==alpha[i]) { // he got it right
            correct++;
            if (echo) playMorseCode(ditdahs, 0);
          } else {
            while (ditdahs != dida && ditdahs!=0) {  // give them the correct ans, & do so until they get it right
              brt[7] = brt[8] = true;
              buZZ(500,1000);
              brt[7] = brt[8] = false;  refreshDisp();
              playMorseCode(dida, 0);
              ditdahs = fetchMorseChar(true);
              scanBtns();
              if (btn4) return;
            }
            if (echo) playMorseCode(ditdahs, 0);
          }
          clearDisp();
//            delay(500);  // ? whats advisable here
          break;
      }
      if(btn3Pressed) echo = !echo;
      if (ESC || btn4) return;
    }
  }

  Serial.print("\ncorrect: "); Serial.print(correct); 
  Serial.print("  out of  "); Serial.println(26); 
  showLevel(correct);
  refreshWait(3000); clearDisp();
}

// -----------------------   
void Morse_AB_test() {
//  const char alpha[] = "ZABCDEFGHIJKLMNOPQRSTUVWXYZA";
  byte clist[28], ri[28];
  int i,ci;
  int correct;
  char cinput;
  byte dida, ditdahs;
  byte ntests=12;  // how many to do

  while (1) {
    delay(1000);
    correct=0;
    ranIndexes(ri, 26); // get randomized set of indexes, to chars of alphabet
          
    for (int i=1; i<=ntests; i++) {
      while(Btn) scanBtns();  // wait out old btn presses
      ci = ri[i];
      dida = Ascii2dida(alpha[ci]);
      set_didaStr(dida, didaStr);
      playMorseCode(dida, 0);              // ---- sound one out
      Serial.println();
      Serial.print(didaStr);  Serial.print(" ");  Serial.print(alpha[ci]);
      ditdahs = fetchMorseChar(true);  // ----  get their response
      if (btn4) return;
      if (btn3) {
        i--;        // replay the prompt char
        delay(750);
        continue;  // go back and do it again
      }
      cinput = dida2Ascii(ditdahs);
      if (cinput==alpha[ci+1]) { // he got it right
        correct++;
      } else {
        while (cinput!=alpha[ci+1]) {  // rePrompt them, until they get it right
          if (ditdahs!=0) {
            brt[7] = brt[8] = true;
            buZZ(500,1000);
            brt[7] = brt[8] = false; refreshDisp();
          }
          if (btn3) delay(750);
          while(Btn) scanBtns();
          playMorseCode(dida, 0);
          ditdahs = fetchMorseChar(true);
          if (btn4) return;
          cinput = dida2Ascii(ditdahs);
        }
      }
      clearDisp();
      delay(750);    
      if (ESC || btn4) return;
    }
    Serial.print("\ncorrect: "); Serial.print(correct); 
    Serial.print("  out of  "); Serial.println(ntests); 
    showLevel(correct);
    refreshWait(3000); clearDisp();
  }
}


// -----------------------   
void Relay_exercise() {
  const char SpChar[] = "01234567890.,?/'-.,?/'-:\"+!=@$)&\n";
  char dd[10];  // stores ditdah given out
  byte ci, li, wlen, numW, mini, maxi;
  int i,p,dl;
  static int opt=1;
  char dida;
  char c, cinput;
  byte correct;
  static int ditPace;
  static bool adjd=false;
  float wpm;
  
  if (!adjd) ditPace = 5*ditPeriod/6;  // use 5/6 his current dit period, play at 1.2x keying speed
  // note: once manually overriden, new speed will get used
  wpm=((1000.0*60.0) / (240.0*ditPace/5.0)); 
  Serial.print(" Audio out rate to be used: "); Serial.println(wpm);
  
  opt=get_Option(5,opt);
  if (ESC || opt==0) return;
  mini=1;  // default
  //  determin the char set to be used
  if      (opt==1) {mini=1; maxi=26;}  // Alpha
  else if (opt==2) {mini=1; maxi=51;}  // Alpha + Nums + punctation . thru + 
  else if (opt==3) {mini=27; maxi=51;} // Nums + punctation . thru +
  else if (opt==4) {mini=37; maxi=58;} // just punctation-special chars

  wlen = (opt==5)? 5:4;
  numW = (opt==5)? 4:3;

  while(!btn4) {
    correct=0;
    clearDisp();
    while(btnDown()) refreshWait(1);

    // ------  wait for two second, mean while btn2&3 can be used to adjust speed
    dl=2000;  // delay for 2secs
    while(dl--) {
      scanBtns(); refreshWait(1);
      if (btn4) return;
      if (btn2Pressed || btn3Pressed) {
        if (btn3) ditPace = (5*ditPace+3)/6;  // pickup the pace 20% faster 
        else      ditPace = (6*ditPace+2)/5;  // slow down the pace 20% slower
        adjd=true;
        wpm=((1000.0*60.0) / (240.0*ditPace/5.0)); 
        Serial.print(" Audio out rate: "); Serial.println(wpm);
      }
      if (BtnPressed) {
        dl=3000;  // extend waiting
        playMorseCode(Ascii2dida('P'),ditPace);  // give a 'P' for reference
      }
    }
  
    for (int w=1; w<=numW; w++) {
      if (opt==5) maxi = (random(3)==0)? 26 : 42;
      for (li=0; li<wlen; li++) { // generate random letters of a psuedo word
        ci = random(mini,maxi+1); // get the character form the set he selected
        if (opt==5 && li==4) ci = random(37,58); // add an extra spc. char
        if (ci<27) {
          c = alpha[ci];
          Serial.print(c);
          dida = Ascii2dida(c);
        } else {
          i=ci-26;
          c = SpChar[i];
          Serial.print(c);
          if (i>16) i-=6;  // adjust index to skip over repeated chars in SpChar[] 
          dida = nonAlpha_dida[i];
        }
        dd[li]=dida;
  
        playMorseCode(dida, ditPace);
        if (btn4) return;
        if (ditPace>120)  delay(2*ditPace);  // std timing between char pause
        else delay(2*ditPace*farns);  // give students +time to recallet with farnsworth
        if (farns==1) delay(3*ditPace);  // still give +time to recallet std speed chars, they just heard
  
        set_didaStr(dida, didaStr);
        Serial.print(" "); Serial.print(didaStr); Serial.print("    ");
      }




      
      Serial.println();
      while (!(btn1 || btn2)) {
        refreshWait(1);
        scanBtns();
        if (btn4) return;
        if (btn3Pressed) {
          delay(1000);           
          for (li=0; li<wlen; li++) {   // replay the 'word'
            playMorseCode(dd[li], ditPace);
            delay(2*ditPace*farns);
          }
        }
      }
      for (li=0; li<wlen; li++) { // get their reponse to the psuedo word given
        dida = fetchMorseChar(true);
        if (ESC || btn4) return;
        Serial.print("  ");
        if (dida==dd[li]) correct++;
        else buZZ(150,100);
      }
      Serial.println();
      delay(1500);
    }
    Serial.print("\ncorrect: "); Serial.print(correct); 
    Serial.print("  out of ");  Serial.println(wlen*numW);
    if (correct==0) correct=1; // credit for participation
    showLevel(correct);
    refreshWait(2000); //clearDisp();
  }
}

// -----------------------   
void Tx_10Word_test() {
  char line[62];
  char wrd[10];
  byte ci, si, ri, wlen, errs, ccnt;
  int i, p, ln;
  static int lln;
  byte ditdahs;
  char cinput;
  float wpm;
  unsigned long t;
  const int hiSpd=11;
  
  wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0));
  //Serial.print("\nStarting char WPM rate: "); Serial.println(wpm);
  if (wpm>=hiSpd) {beep(45,30);  beep(50);}  // let known punc. will be included
  do {
    ln = random(4); // randomly select line 1-4
  }  while (ln==lln);  // don't repeat same as last line
  lln=ln;
  for (i=0; i<62; i++) {line[i] = wordSet[ln][i];}  // crude string copy
  Serial.print(" Line #");Serial.println(ln+1); 
  Serial.println(wordSet[ln]);
  set_LEDs(0b11 << (2*ln),2); // light LED pair 1-4
  ci=0;
  ccnt=0;
  t0=0;
  errs=0;
  for (int w=0; w<10; w++) {
    si=ci;
    if (wpm<hiSpd) {
      while (toupper(line[ci])>='A' && toupper(line[ci])<='Z') {ci++;}  // accept only alphabet chars
    } else {
      while (line[ci]!=0 && line[ci]!=' ') {ci++;}      // if their speed is > 11 take puncuation too
    }
    for (i=0; (si+i)<ci; i++) {wrd[i]= line[si+i];}
    wlen =i;
    if (line[ci]!=' ') ci++;
    wrd[i]=' '; wrd[i+1]=0;
    Serial.println(); Serial.print(wrd);
    dimFlash= 1+ (w & 0x07);
    refreshWait(0);
    ci++;

    // --------  get their response
    delay(100);
    ri=0;
///    while (ri++<9) {
    for (ri=0; wrd[ri]!=0 && wrd[ri]!=' '; ri++) {
      while(!Btn) {refreshWait(1); scanBtns();} // wait for button press
      dimFlash=litLED=0;
      refreshDisp();
      ditdahs = fetchMorseChar(true);
      if (t0==0) t0=millis();
      if (ESC || btn4) return;
      litLED=1+ (w & 0x07);
      ccnt++;
      cinput = dida2Ascii(ditdahs);
      if (cinput!=toupper(wrd[ri])) { // did they get it right
        errs++;
      }
      refreshWait(1);   
      p=0;

      if ( wrd[ri+1]==0 || wrd[ri+1]==' ') continue;
      while (!(btn1||btn2)) {  // wait for start of next letter (or end of word)
        scanBtns();
        if (ESC || btn4) return;
        refreshWait(1);
///        if (++p>(5*ditPeriod)) { // give 3+5 more than 7p (3+4p) before thinking their done, as this is a test
        if (++p>(14*ditPeriod)) {  //
//        ri=10;
          errs++;
          ccnt++; // adj count so that WPM calc is not ill effected
          wrd[ri+1]=0;
          break;   // done here, go get next word
        }
      }
    }
    litLED=0; refreshDisp();
    beep(30);  // going on to next word
  }
  
  t=millis()-t0;
  wpm=((1000.0*60.0) / (240.0*ditPeriod/5.0));
  Serial.print("\n Current Keying ditdah rate in WPM: "); Serial.print(wpm);

  wpm=((ccnt-1+2)/5.0)/(t/1000.0/60.0); // +2 is adj for alowing extra time between letters & words
  Serial.print("  over all WPM: "); Serial.print(wpm);
  showLevel(wpm+0.5);
  refreshWait(3000); clearDisp();
  if (errs>0) {
    Serial.print("   Errors made: "); Serial.println(errs); 
    buZZ(150);
    showLevel(errs);
    refreshWait(3000); clearDisp();
  } else {
    Serial.println("    No errors were detected.");
  }
}

// -----------------------   
void Rx_10Word_test() {
  char line[100];  // has extra space for a random word set + added nums,punc,Spc.
  char wrd[10];
  byte ci, si, ri, li, wn, wlen, errs, ccnt;
  int i,p;
  byte ditdahs;
  char c, cinput;
  unsigned long t;
  long wpm;
  static int opt=2;
  int ditPace;
  int ln, L;
  static int lln, llln;  // track the Last Line Nums.
  int lmax;
  int ff;
  
  //   #s adjusted to compensate for processing overhead
  // set dit period to get WPM per Opt 1:8 2:12 3:16 4:20 5:24 6:28 7:32 8:36
  opt=get_Option(8, opt);   // ------ get speed Option
  if (ESC || opt==0) return;
  ditPace=300/(opt+1);   // empirically devined
  ff = farns;     // set local Farnsworth Factor
  if (ff!=1 && opt<3) ff = (ff+1)/2;  // with opt 1&2, limit Farnsworth Factor
  
  wpm=4*opt+4;
  Serial.print(" Audio out, approximate char speed in WPM: "); Serial.println(wpm);
  clearDisp();
  refreshWait(200);
    
//  lmax = (opt>2)? 8 : 7;
  do {   // randomly select line 1-6 or the 7th case of a random select of 10 words
    ln = random(8)+1;
  }  while (ln==lln || ln==llln);  // don't repeat same as last line
  llln=lln; lln=ln;
  if (opt<3 && ln==8) ln=7; // ...only with speeds over 10wpm: 8th to include numbers, punctuation and spc. chars

  brightOne = ln;
  refreshDisp();
  Serial.print(" Line #");Serial.println(ln); 
                     
  if (ln<7) {
    for (i=0; i<62; i++) line[i] = wordSet[ln-1][i];  // crude string copy
  } else {      // ---------------- create a line of random words
    li=0;
    for (int n=0; n<10; n++) {
      L = random(6);
      wn = random(1,11);
      ci=0;
      for (int w=0; w!=wn; w++) {
        si=ci;
        // -----  skip to to end of word
        while (toupper(wordSet[L][ci])>='A' && toupper(wordSet[L][ci])<='Z') {ci++;}  // accept only alphabet chars
        if (wordSet[L][ci]!=' ') ci++;
        wlen=ci-si;
        ci++;
      }
      if (ln==8 && random(2)) line[li++]= nonAlpha[random(1,26)];   // add in #s + spc if appropriate
      for (i=0; i<wlen; i++) {line[li+i]= wordSet[L][si+i];}
      if (ln==8 && random(2)) line[i+li++]= nonAlpha[random(1,26)]; //
      line[li+i]=' ';
      li=li+i+1;
    }
    line[li]=0;
  }
  // Serial.println(line);
  refreshWait(1500);
  clearDisp();
  stableBtns();
  if (btn4) return;
  
  ci=0; ccnt=0;
  t0=millis();
  for (int w=0; w<10; w++) {  // ------ Output the Morse code for the Rx test '10 word' line
    si=ci;
    while (line[ci]!=0 && line[ci]!=' ') ci++;  // find the end of this 'word'
    for (i=0; (si+i)<ci; i++) {wrd[i]= line[si+i];}
    wlen =i;
    wrd[i]=line[ci]; wrd[i+1]=0;
    if (line[ci]!=' ' && line[ci]!=0) {
      wlen++;
      if (opt<3) wrd[i]=' ';
      ci++;  // skip extra in the source
    }
    ci++;
      
    // -------------   Output the ditdahs for letters of current word
    for (ri=0; ri<wlen; ri++) {
      c = toupper(wrd[ri]);
      Serial.print(c);
      if (c!=' ') {
        playMorseCode(Ascii2dida(c), ditPace);
        delay(ff*ditPace-ditPace);  // Adjust end of char space (there alreay was 1 ditPace
        ccnt++;
      }
      // wait for 2 additional p-periods between chars (for 3 total)
      delay(2*(ff*ditPace));
      if (ff==1) {  // ----- during std timing
        // ----- still give students just a little extra time to discern
        delay(ditPace);   // +1p for 4 instead of 3p, may be more real word like
      }
      
      stableBtns();
      if (ESC || btn4) return;
      if (Btn) {                        // go into Pause state
        set_LEDs(0x30, 3);
        while (Btn) scanBtns();
        while (!BtnPressed) scanBtns(); // wait for next press to continue
        set_LEDs(0x30, 0);
        delay(750);
      }
    }
    Serial.print(' ');
///    delay((7-3)*ditPace);
    delay((7-3)*ditPace*ff);

    litLED=0;
  }
  Serial.println();
  
  if (DEBUG) {
    t=millis()-t0;
    wpm=((1000.0*60.0) / (240.0*ditPace/5.0));
    Serial.print("\n Character ditdah speed in WPM: "); Serial.print(wpm);
    wpm=(5*(ccnt*12+9*4)/240)/(t/1000.0/60.0);  // 5 wpm for each 240*p
    Serial.print("   ~ Overall speed Words/Min: "); Serial.println(wpm);
  }
}

// -----------------------   
void printTest_Chr_Ditdah() {
  const char SpChar[] = "01234567890.,?/'-:\"+!=@$)&\n";
  byte ci, i, dida;
  char c;
  char hex[3];
  
//Serial.print(" nonAlpha 1 - "); Serial.print(sizeof(nonAlpha)-1);
//Serial.print("   nonAlpha_dida 1 -1 "); Serial.println(sizeof(nonAlpha_dida)-1);

  for (ci=1; ci<=52; ci++) {
    Serial.print(ci); Serial.print("  ");
    if (ci<27) {
      c = alpha[ci];
      Serial.print(c);
      dida = Ascii2dida(c);
    } else {
      i=ci-26;
      c = SpChar[i];
      Serial.print(c);
      //if (i>14) i-=4;  // adjust index to skip over repeated chars in SpChar[] 
      dida = nonAlpha_dida[i];
    }
    if (btnDown()) return;
    
    // check conversion utilities
    sprintf(hex, "%02X",Ascii2dida(c));
    Serial.print(dida2Ascii(Ascii2dida(c))); Serial.print(" : "); Serial.print(hex);

    playMorseCode(dida, 0);
    set_didaStr(dida, didaStr);
    Serial.print(" = ");
    Serial.print(didaStr); Serial.println("  ");
    delay(2*ditPeriod*farns);
    if (btnDown()) return;
  }
}

// ================== Local Support Utility Functions =================

// -----------------------   
void dit_dah(unsigned short int delayms){  // Used for ditdah audio while user has key pressed
                                          // higher freq & longer duty cycle for piezo
  delayms = delayms - (delayms/5);
  for (int i=0; i<delayms; i++) {
    digitalWrite ( BEEPPIN, BUZZER_ON); 
    delayMicroseconds(750);
    digitalWrite ( BEEPPIN, BUZZER_OFF);
    delayMicroseconds(500);
  }
  digitalWrite ( BEEPPIN, BUZZER_OFF);;
}

// -----------------------   
void ditAudio(unsigned short int delayms){  //  over duty cycle adjusted for freq.
                                  // ON duration extra needed for good piezo sound
//  delayms = delayms + (delayms/2);  // for use with 400/266
  delayms = delayms - (delayms/5);  // for use with 700/550
  for (int i=0; i<delayms; i++) {         // wait for 'delayms' millisecs
    digitalWrite ( BEEPPIN, BUZZER_ON); 
    delayMicroseconds(750);    // tried: 700 600 400 456
    digitalWrite ( BEEPPIN, BUZZER_OFF);
    delayMicroseconds(500);   // tried: 550 400 266 210
    refreshDisp();  // just spent ~1 ms, lets refreshDisp now
  }
  digitalWrite ( BEEPPIN, BUZZER_OFF);;
}

// -----------------------   
void dahAudio(unsigned short int delayms){  // lower freq ( w/ longer ON duty cycle for Piezo)
  delayms = (2*delayms)/3;
  for (int i=0; i<delayms; i++) {         // wait for delayms ms
    digitalWrite ( BEEPPIN, BUZZER_ON); 
    delayMicroseconds(900);    
    digitalWrite ( BEEPPIN, BUZZER_OFF);
    delayMicroseconds(600);
    refreshDisp();  // just spent ~1 ms, lets refreshDisp now
  }
  digitalWrite ( BEEPPIN, BUZZER_OFF);;
}

// -----------------------   
void playMorseCode(byte ditdahs, int pace) {
  // all MSBits that are the same are discarded, whats left is 1=dit 0=dah, MSbit is first in time, LSB last
  short int ditPeriod;
  byte i=1;
  byte bitState;

  ditPeriod = pAveraging/8;    // set according to their current pace
  ditPeriod = 8*ditPeriod/10;  // but 80% timing, for 1.25 the speed
  if (pace) ditPeriod=pace;    // unless passed in

  if (ditdahs==0) return;
  while ((ditdahs<<i & 0x80) == (ditdahs & 0x80)) i++;
  for (int b=i; b<8; b++) {
    if (ditdahs<<b & 0x80) {  // this bit is a 'dit'
      brightOne=7;
      ditAudio(ditPeriod);
      brightOne=0;
      refreshWait(ditPeriod);      
    } else {  // its a dah
      dim[7]=dim[8]=true;
      dahAudio(3*ditPeriod);
      dim[7]=dim[8]=false;
      refreshWait(ditPeriod);
    }
    scanBtns();  // needed to keep button state upto date (for btn4 esc)
  }
///  refreshWait(2*ditPeriod);
}

// -----------------------   
void playStr(char *wrd) {          // ment for messsage announcements
  float ff=(farns==1)? 1:farns/2;  // so OK to go easy on the farnsworth effect
  char c;
  Serial.println(wrd);
  for (int i=0; i<strlen(wrd); i++) {
    playMorseCode(Ascii2dida(wrd[i]), 0);
    delay(2*ditPeriod*ff);
  }
}

// ================ Generally Useful Support Utility Functions =============== 
void showLevel(byte level) {
  int n;
  clearDisp();
  refreshWait(250);
  for (int i=0; i<level; i++) {
    beep(10);
    n = (i%nleds)+1;  // n wraps around as i exceeds nleds
    dim[n]=true;
    brightOne=n;
    refreshWait(100);
  }
}

// -----------------------   
byte get_Option(byte optCnt, int dflt) {  // optCnt is the max # option, dflt is the default value
  clearDisp();
  for (int i=1; i<=optCnt; i++) {dim[i]=true;}
  flashLed=dflt;
  while(btnDown()) ;  // wait for button release
  while(! btn1) {
      refreshWait(1);
      if (btnPressed()) {
        if (btn2Pressed) flashLed = (flashLed==0)?  optCnt : flashLed-1;  // -- move left Allow 0:debugging
        if (btn3Pressed) flashLed = (flashLed%optCnt) + 1;              // ++ moves right
        //if (btn4Pressed) flashLed = (flashLed%optCnt) + 1; // allow most play with only 2 buttons (1&4)
      }
      if (ESC) return(0);
  }
  return(flashLed);
}

// --------------------------
// produce a requested number of randomly sorted indexes
void ranIndexes(byte rIndexes[], byte cnt) {  
  char wi[80];  // working set of indexes
  int i, indx;
  byte tmp;
  
  for (int c=1; c<=cnt; c++) wi[c]=c; // preset the collection of indexes
  
  for (int s=cnt; s>0; s--) {  // simple fast random sort
    i = random(1,s);
    rIndexes[s]=wi[i];
    wi[i]=wi[s];
  }
  // ensure there are not any consecutive entries, not for randomness but AB_exersice desired behavior
  for (i=1; i<cnt; i++) {
    if (rIndexes[i]+1==rIndexes[i+1]) {
      tmp = rIndexes[i];
      rIndexes[i]=rIndexes[i+1];
      rIndexes[i+1]=tmp;
    }
    //Serial.print(rIndexes[i]); Serial.print(' ');
  }
  //Serial.println(rIndexes[i]);
}

// -----------------------   
bool btnDown() {
  scanBtns();
  return (Btn!=0);
}

// -----------------------   
void stableBtns() { // get a stable state of the buttons post debouncing
  refreshWait(1);
  scanBtns();
  while(!btnStable) {
    refreshWait(1);
    scanBtns();
  }
}
